# from setup import _game_name

# base_module = __import__(f"src.Base.{_game_name}.env", fromlist=["*"])
# from src.render_template import Render as __Render, import_files as __import_files
# from numba.core.errors import NumbaPendingDeprecationWarning as __NumbaPendingDeprecationWarning
# import warnings as __warnings
# __warnings.simplefilter("ignore", __NumbaPendingDeprecationWarning)


# __import_files(_game_name)

# __env = base_module

# bot_lv0 = __env.bot_lv0
# getValidActions = __env.getValidActions
# getActionSize = __env.getActionSize
# getAgentSize = __env.getAgentSize
# getStateSize = __env.getStateSize
# getReward = __env.getReward
# run = __env.run

# def render(Agent = 'human',
#            per_data: any = [0],
#            level: int = 0,
#             *args,
#             max_temp_frame=100):
#     list_agent, list_data = __env.load_agent(level, *args)

#     if "__render" not in globals():
#         global __render
#         __render = __Render(Agent, per_data, list_agent, list_data, max_temp_frame)
#     else:
#         __render.__init__(Agent, per_data, list_agent, list_data, max_temp_frame)

#     return __render.render()


# def get_data_from_visualized_match():
#     if "__render" not in globals():
#         print("Nothing to get, visualize the match before running this function")
#         return None

#     return {
#         "history_state": __render.history_state,
#         "history_action": __render.history_action,
#     }



import numpy as np
from PIL import Image, ImageDraw, ImageEnhance, ImageFont

from src.Base.SushiGo import env as _env
from env import SHORT_PATH

IMG_PATH = SHORT_PATH + "src/Base/SushiGo/images/"
BG_SIZE = (1680, 720)
CARD_SIZE = (90, 130)


action_description = {
    0: "Tempura",
    1: "Sashimi",
    2: "Dumpling",
    3: "1 Maki Roll",
    4: "2 Maki Roll",
    5: "3 Maki Roll",
    6: "Salmon Nigiri",
    7: "Squid Nigiri",
    8: "Egg Nigiri",
    9: "Pudding",
    10: "Wasabi",
    11: "Chopsticks",
    12: "Use Chopsticks",
    13: "End Turn",
}


class Env_components:
    def __init__(self, env, winner, list_other, turn, round) -> None:
        self.env = env
        self.list_action = np.full((5, 3), 13)
        self.winner = winner
        self.list_other = list_other
        self.idx = -1
        self.turn = turn
        self.round = round
        self.count = 0
        self.state = None


def get_description(action):
    if action < 0 or action >= _env.getActionSize():
        return ""

    return f"{action_description[action]}"


def get_env_components():
    env = _env.initEnv(5)
    winner = _env.winner_victory(env)
    list_other = np.array([-1, 1, 2, 3, 4])
    np.random.shuffle(list_other)
    while list_other[-1] == -1:
        #  print(list_other, 'change')
        np.random.shuffle(list_other)
    #  print(list_other)
    turn = env[1]
    round = env[0] - 1
    env_components = Env_components(env, winner, list_other, turn, round)
    return env_components


def get_main_player_state(
    env_components: Env_components, list_agent, list_data, action=None
):
    amount_player = 5
    if not action is None:
        env_components.list_action[env_components.idx][env_components.count] = action
        # print(env_components.turn, env_components.round, env_components.idx, env_components.count)
        # print(env_components.list_action)

    turn = env_components.env[1]
    check_end_game = False

    check_break = True
    while turn <= 7 * 3:
        turn = env_components.env[1]
        round = env_components.env[0] - 1
        env_components.turn = turn
        env_components.round = round
        if env_components.idx == 4:
            env_components.idx = -1
        #  if turn % 7 == 0:
        #      env_components.list_action = np.full((amount_player, 3), 13)

        check_use_chopsticks = False
        if env_components.list_other[env_components.idx] == -1:
            if (
                12 in env_components.list_action[env_components.idx]
                and env_components.count != 2
            ):
                check_use_chopsticks = True
                env_components.idx -= 1
            if env_components.count == 2:
                env_components.count = 0
                env_components.state = None

        for idx in range(env_components.idx + 1, amount_player):
            env_components.idx = idx
            player_state = _env.getAgentState(env_components.env, idx)
            count = env_components.count
            while player_state[-1] + player_state[-2] > 0:
                if env_components.list_other[idx] == -1:
                    check_break = False
                    if check_use_chopsticks:
                        # print('Use chopsticks')
                        # print(player_state[14:28])
                        if action == 12:
                            env_components.state = player_state
                        player_state = _env.test_action(env_components.state, action)
                        # print(player_state[14:28])
                        env_components.state = player_state
                        env_components.count += 1

                    break
                agent = list_agent[env_components.list_other[idx] - 1]
                data = list_data[env_components.list_other[idx] - 1]
                action, data = agent(player_state, data)
                env_components.list_action[idx][count] = action
                count += 1
                env_components.count = count
                player_state = _env.test_action(player_state, action)

            #  env_components.idx += 1
            if check_break == False:
                break
            env_components.count = 0
        if check_break == False:
            break
        env_components.env = _env.stepEnv(
            env_components.env, env_components.list_action, amount_player, turn, round
        )
        env_components.list_action = np.full((amount_player, 3), 13)
        if turn % 7 == 0:
            env_components.env = _env.calculator_score(
                env_components.env, amount_player
            )
            if env_components.env[0] < 3:
                env_components.env[0] += 1
                env_components.env = _env.reset_card_player(env_components.env)
        if turn == 7 * 3:
            #  print('Tính pudding')
            env_components.env = _env.calculator_pudding(
                env_components.env, amount_player
            )
        if turn <= 7 * 3:
            env_components.env[1] += 1
            #  print(env_components.env[1], turn)

    if check_break == True:
        check_end_game = True
    env_components.winner = _env.winner_victory(env_components.env)

    if check_end_game == False:
        win = -1
    else:
        my_idx = np.where(env_components.list_other == -1)[0][0]
        env = env_components.env.copy()

        player_state = _env.getAgentState(env_components.env, env_components.idx)
        #  print(player_state[:2], player_state[14:28])
        #  print(env_components.list_other, env_components.idx)
        if my_idx in env_components.winner:
            win = 1
        else:
            win = 0

        #  Chạy turn cuối cho 3 bot hệ thống
        for idx in range(amount_player):
            if idx != my_idx:
                _state = _env.getAgentState(env, idx)
                agent = list_agent[env_components.list_other[idx] - 1]
                data = list_data[env_components.list_other[idx] - 1]
                action, data = agent(_state, data)

    return win, player_state, env_components


class Sprites:
    def __init__(self) -> None:
        self.background = Image.open(IMG_PATH + "bg.jpg").resize(BG_SIZE)
        card_values = [
            "temura",
            "sashimi",
            "dumpling",
            "1maki",
            "2maki",
            "3maki",
            "salmon",
            "squid",
            "egg",
            "pudding",
            "wasabi",
            "chopsticks",
        ]
        self.cards = []
        for value in card_values:
            self.cards.append(Image.open(IMG_PATH + f"{value}.jpg").resize(CARD_SIZE))


class Params:
    def __init__(self) -> None:
        self.center_card_x = BG_SIZE[0] * 0.5
        self.center_card_y = (BG_SIZE[1] - CARD_SIZE[1]) * 0.35

        x_0 = BG_SIZE[0] * 0.12
        x_1 = BG_SIZE[0] * 0.8
        y_0 = 0.1 * BG_SIZE[1] - 0.25 * CARD_SIZE[1]
        y_1 = 0.9 * BG_SIZE[1] - 0.75 * CARD_SIZE[1]
        x_center = BG_SIZE[0] * 0.45
        y_center = 0.9 * BG_SIZE[1] - 0.75 * CARD_SIZE[1]
        self.list_coords_1 = [
            (x_center, y_center),
            (x_0, y_1),
            (x_1, y_1),
            (x_1, y_0),
            (x_0, y_0),
        ]

        self.score_coords = [
            (x_center, y_center * 0.9),
            (x_0, y_1 * 0.9),
            (x_1, y_1 * 0.9),
            (x_1, y_0 * 6),
            (x_0, y_0 * 6),
        ]


params = Params()
sprites = Sprites()


def draw_cards(bg, cards, s, y):
    y = round(y)
    id_card = 1
    for card in range(12):
        total_cards = cards[card]
        while total_cards > 0:
            bg.paste(sprites.cards[card], (round(s + _d_ * id_card), y))
            total_cards -= 1
            id_card += 1


_d_ = CARD_SIZE[0] * 0.8

font_ = ImageFont.truetype("src/ImageFonts/FreeMonoBoldOblique.ttf", 60)


def get_state_image(state=None):
    background = sprites.background.copy()
    if state is None:
        return background
    else:
        n = np.sum(state[2:14])
        w = CARD_SIZE[0] + _d_ * (n - 1)
        s = params.center_card_x - 0.5 * w
        draw_cards(background, state[2:14], s, params.center_card_y)

    for k in range(5):
        list_cards_played = state[14 * (k + 1) + 2 : 14 * (k + 2)]
        score = int(state[14 * (k + 1)])
        count_puding = int(state[14 * (k + 1) + 1])
        n = np.sum(list_cards_played)
        w = CARD_SIZE[0] + _d_ * (n - 1)
        s = params.list_coords_1[k][0] - 0.5 * w
        draw_cards(background, list_cards_played, s, params.list_coords_1[k][1])

        ImageDraw.Draw(background).text(
            params.score_coords[k],
            f"{score}|{count_puding}",
            fill=(255, 255, 255),
            anchor="mm",
            font=font_,
        )
    return background



import sys
import numpy as np
from numba import njit
import env

game_name = sys.argv[1]
env.make(game_name)


from numba.typed import List


# Agent function
def DataAgent():
    per = List(
        [
            np.random.choice(
                np.arange(env.getActionSize()), size=env.getActionSize(), replace=False
            )
            * 1.0,
            np.zeros(env.getActionSize()),
        ]
    )
    return per


def convert_to_save(perData):
    if type(perData) == np.ndarray:
        raise Exception("Data này đã được convert rồi.")
    return perData[1]


def convert_to_test(perData):
    return perData


@njit()
def Train(state, per):
    actions = env.getValidActions(state)
    output = actions * per[0] + actions
    action = np.argmax(output)
    win = env.getReward(state)
    if win == 1:
        per[1] += per[0]
    if win == 0:
        np.random.shuffle(per[0])
    return action, per


@njit()
def Test(state, per):
    actions = env.getValidActions(state)
    output = per * actions + actions
    action = np.argmax(output)
    return action, per



import numpy as np
from PIL import Image, ImageEnhance

from src.Base.TLMN import env as _env
from env import SHORT_PATH

IMG_PATH = SHORT_PATH + "src/Base/TLMN/images/"
BG_SIZE = (1680, 720)
CARD_SIZE = (80, 112)


class Sprites:
    def __init__(self) -> None:
        self.background = Image.open(IMG_PATH + "background.png").resize(BG_SIZE)
        card_values = ["3", "4", "5", "6", "7", "8", "9", "10", "J", "Q", "K", "A", "2"]
        card_suits = ["Spade", "Club", "Diamond", "Heart"]
        self.cards = []
        for value in card_values:
            for suit in card_suits:
                self.cards.append(
                    Image.open(IMG_PATH + f"{value}-{suit}.png").resize(CARD_SIZE)
                )

        self.card_back = Image.open(IMG_PATH + "Card_back.png").resize(CARD_SIZE)
        self.faded_card_back = self.card_back.copy()
        br = ImageEnhance.Brightness(self.faded_card_back)
        self.faded_card_back = br.enhance(0.5)
        ct = ImageEnhance.Contrast(self.faded_card_back)
        self.faded_card_back = ct.enhance(0.5)


sprites = Sprites()


class Params:
    def __init__(self) -> None:
        self.center_card_x = BG_SIZE[0] * 0.5
        self.center_card_y = (BG_SIZE[1] - CARD_SIZE[1]) * 0.5
        self.list_coords_0 = [
            (self.center_card_x, 0.92 * BG_SIZE[1] - CARD_SIZE[1]),
            (0.82 * BG_SIZE[0], self.center_card_y),
            (self.center_card_x, 0.08 * BG_SIZE[1]),
            (0.18 * BG_SIZE[0], self.center_card_y),
        ]

        x_0 = BG_SIZE[0] * 0.32
        x_1 = BG_SIZE[0] * 0.68
        y_0 = 0.2 * BG_SIZE[1] - 0.25 * CARD_SIZE[1]
        y_1 = 0.8 * BG_SIZE[1] - 0.75 * CARD_SIZE[1]
        self.list_coords_1 = [(x_0, y_1), (x_1, y_1), (x_1, y_0), (x_0, y_0)]


params = Params()


_d_ = CARD_SIZE[0] * 0.2


def draw_cards(bg, cards, s, y, back=False, faded=False):
    n = cards.shape[0]
    y = round(y)
    if back:
        if faded:
            im = sprites.faded_card_back
        else:
            im = sprites.card_back

        for i in range(n):
            bg.paste(im, (round(s + _d_ * i), y))
    else:
        for i in range(n):
            bg.paste(sprites.cards[cards[i]], (round(s + _d_ * i), y))


def get_state_image(state=None):
    background = sprites.background.copy()
    if state is None:
        return background

    my_cards = np.where(state[0:52])[0]
    n = my_cards.shape[0]
    w = CARD_SIZE[0] + _d_ * (n - 1)
    s = params.list_coords_0[0][0] - 0.5 * w
    draw_cards(background, my_cards, s, params.list_coords_0[0][1])

    for k in range(1, 4):
        faded = not state[103 + k]
        n = state[106 + k]
        w = CARD_SIZE[0] + _d_ * (n - 1)
        if k == 1:
            s = params.list_coords_0[1][0] - w
        elif k == 2:
            s = params.list_coords_0[2][0] - 0.5 * w
        else:
            s = params.list_coords_0[3][0]

        draw_cards(
            background, np.full(int(n), 0), s, params.list_coords_0[k][1], True, faded
        )

    cards_played = np.where(state[52:104])[0]
    cur_cards = np.where(state[113:165])[0]
    cards_played = np.setdiff1d(cards_played, cur_cards)

    n = cur_cards.shape[0]
    w = CARD_SIZE[0] + _d_ * (n - 1)
    s = params.center_card_x - 0.5 * w
    draw_cards(background, cur_cards, s, params.center_card_y)

    list_cards_played = np.array_split(cards_played, 4)
    for k in range(4):
        n = list_cards_played[k].shape[0]
        w = CARD_SIZE[0] + _d_ * (n - 1)
        s = params.list_coords_1[k][0] - 0.5 * w
        draw_cards(background, list_cards_played[k], s, params.list_coords_1[k][1])

    return background


hand_name = {
    1: "Single",
    2: "Pair",
    3: "Three of a kind",
    4: "Four of a kind",
    5: "3-card straight",
    6: "4-card straight",
    7: "5-card straight",
    8: "6-card straight",
    9: "7-card straight",
    10: "8-card straight",
    11: "9-card straight",
    12: "10-card straight",
    13: "11-card straight",
    14: "3-pair straight",
    15: "4-pair straight",
}

card_name = []
card_values = ["3", "4", "5", "6", "7", "8", "9", "10", "J", "Q", "K", "A", "2"]
card_suits = ["Spade", "Club", "Diamond", "Heart"]
for value in card_values:
    for suit in card_suits:
        card_name.append(f"{value}-{suit}")


def get_description(action):
    if action < 0 or action >= _env.getActionSize():
        return ""

    if action == 0:
        return "Skip"

    hand, card = _env._ACTIONS_[action]
    return f"{hand_name[hand]}, high card {card_name[card]}"


class Env_components:
    def __init__(self, env, cur_cards, winner, list_other) -> None:
        self.env = env
        self.cur_cards = cur_cards
        self.winner = winner
        self.list_other = list_other


def get_env_components():
    env = _env.initEnv()
    while not _env.check_player_hand(env):
        env = _env.initEnv()

    cur_cards = np.full(0, 0)
    winner = _env.checkEnded(env)
    list_other = np.array([-1, 1, 2, 3])
    np.random.shuffle(list_other)

    env_components = Env_components(env, cur_cards, winner, list_other)
    return env_components


def get_main_player_state(
    env_components: Env_components, list_agent, list_data, action=None
):
    if not action is None:
        if action != 0:
            env_components.cur_cards = _env.stepEnv(action, env_components.env)
        else:
            _env.stepEnv(action, env_components.env)

    env_components.winner = _env.checkEnded(env_components.env)
    if env_components.winner == -1:
        while True:
            p_idx = env_components.env[52]
            if env_components.list_other[p_idx] == -1:
                break

            state = _env.getAgentState(env_components.env, env_components.cur_cards)
            agent = list_agent[env_components.list_other[p_idx] - 1]
            data = list_data[env_components.list_other[p_idx] - 1]
            action, data = agent(state, data)
            if action != 0:
                env_components.cur_cards = _env.stepEnv(action, env_components.env)
            else:
                _env.stepEnv(action, env_components.env)

            env_components.winner = _env.checkEnded(env_components.env)
            if env_components.winner != -1:
                break

    if env_components.winner == -1:
        state = _env.getAgentState(env_components.env, env_components.cur_cards)
        win = -1
    else:
        my_idx = np.where(env_components.list_other == -1)[0][0]
        env = env_components.env.copy()
        env[52] = my_idx
        env[57] = my_idx
        cur_cards = np.full(0, 0)
        state = _env.getAgentState(env, cur_cards)
        if my_idx == env_components.winner:
            win = 1
        else:
            win = 0

        #  Chạy turn cuối cho 3 bot hệ thống
        for p_idx in range(4):
            if p_idx != my_idx:
                env[52] = p_idx
                env[57] = p_idx
                _state = _env.getAgentState(env, cur_cards)
                agent = list_agent[env_components.list_other[p_idx] - 1]
                data = list_data[env_components.list_other[p_idx] - 1]
                action, data = agent(_state, data)

    return win, state, env_components



import numpy as np
from PIL import Image, ImageDraw, ImageFont

from src.Base.CatanNoExchange import env as _env
from src.Base.CatanNoExchange.env import MAX_TURN_IN_ONE_GAME
from env import SHORT_PATH

IMG_PATH = SHORT_PATH + "src/Base/CatanNoExchange/images/"

RES_NAME = ["lumber", "brick", "wool", "grain", "ore"]


class Sprites:
    def __init__(self) -> None:
        self.background = Image.open(IMG_PATH + "background.png")
        self._background_ = self.background.copy()

        self.tiles = []
        for tile in RES_NAME + ["desert"]:
            tile_img = (
                Image.open(IMG_PATH + f"tile_{tile}.png")
                .resize((133, 152))
                .convert("RGBA")
            )
            self.tiles.append(tile_img)

        self.probs = {}
        for i in range(2, 13):
            if i != 7:
                prob_img = (
                    Image.open(IMG_PATH + f"prob_{i}.png")
                    .resize((55, 55))
                    .convert("RGBA")
                )
                self.probs[i] = prob_img

        self.ports = []
        for port in RES_NAME:
            port_img = (
                Image.open(IMG_PATH + f"port_{port}.png")
                .resize((80, 80))
                .convert("RGBA")
            )
            self.ports.append(port_img)
        else:
            port_img = (
                Image.open(IMG_PATH + "port.png").resize((80, 80)).convert("RGBA")
            )
            self.ports.append(port_img)

        self.robber = (
            Image.open(IMG_PATH + "icon_robber.png").resize((55, 55)).convert("RGBA")
        )

        self.rds = []
        self.rds_60 = []
        self.rds_120 = []
        self.sts = []
        self.cts = []
        for color in ["mysticblue", "gold", "silver", "bronze"]:
            road_img = (
                Image.open(IMG_PATH + f"road_{color}.png")
                .resize((80, 80))
                .convert("RGBA")
            )
            sett_img = (
                Image.open(IMG_PATH + f"settlement_{color}.png")
                .resize((55, 55))
                .convert("RGBA")
            )
            city_img = (
                Image.open(IMG_PATH + f"city_{color}.png")
                .resize((55, 55))
                .convert("RGBA")
            )
            self.rds.append(road_img)
            self.rds_60.append(road_img.rotate(60, expand=True))
            self.rds_120.append(road_img.rotate(120, expand=True))
            self.sts.append(sett_img)
            self.cts.append(city_img)

        self.highlight_circle = (
            Image.open(IMG_PATH + "icon_highlight_circle_white.png")
            .resize((40, 40))
            .convert("RGBA")
        )
        self.icon_check = (
            Image.open(IMG_PATH + "icon_check.png").resize((50, 50)).convert("RGBA")
        )

        self.largest_army = (
            Image.open(IMG_PATH + "icon_largest_army_highlight.png")
            .resize((40, 40))
            .convert("RGBA")
        )
        self.longest_road = (
            Image.open(IMG_PATH + "icon_longest_road_highlight.png")
            .resize((40, 40))
            .convert("RGBA")
        )

        self.dice = {}
        for i in range(1, 7):
            self.dice[i] = (
                Image.open(IMG_PATH + f"dice_{i}.png").resize((60, 60)).convert("RGBA")
            )


sprites = Sprites()


class Params:
    def __init__(self) -> None:
        self.tile_center_pos = [
            (1397, 210),
            (1538, 210),
            (1609, 331),
            (1679, 453),
            (1609, 574),
            (1538, 696),
            (1397, 696),
            (1256, 696),
            (1186, 574),
            (1115, 453),
            (1186, 331),
            (1256, 210),
            (1327, 331),
            (1256, 453),
            (1327, 574),
            (1468, 574),
            (1538, 453),
            (1468, 331),
            (1397, 453),
        ]
        self.port_topleft_pos = [
            (1421, 31),
            (1632, 153),
            (1774, 396),
            (1632, 639),
            (1421, 759),
            (1139, 758),
            (999, 517),
            (999, 274),
            (1139, 31),
        ]
        self.font24 = ImageFont.FreeTypeFont("src/ImageFonts/arial.ttf", 24)
        self.font28 = ImageFont.FreeTypeFont("src/ImageFonts/arial.ttf", 28)
        self.font32 = ImageFont.FreeTypeFont("src/ImageFonts/arial.ttf", 32)
        self.road_center_pos = [
            (1362, 149),
            (1433, 149),
            (1468, 210),
            (1503, 149),
            (1573, 149),
            (1609, 210),
            (1573, 270),
            (1644, 270),
            (1679, 331),
            (1644, 392),
            (1715, 392),
            (1750, 452),
            (1715, 513),
            (1644, 513),
            (1679, 574),
            (1644, 635),
            (1573, 635),
            (1609, 696),
            (1573, 751),
            (1503, 751),
            (1468, 696),
            (1433, 756),
            (1362, 756),
            (1327, 696),
            (1291, 756),
            (1221, 756),
            (1186, 696),
            (1221, 635),
            (1151, 635),
            (1115, 574),
            (1151, 513),
            (1080, 513),
            (1045, 452),
            (1080, 392),
            (1151, 392),
            (1115, 331),
            (1151, 270),
            (1221, 270),
            (1186, 210),
            (1221, 149),
            (1291, 149),
            (1327, 210),
            (1291, 270),
            (1256, 331),
            (1291, 392),
            (1221, 392),
            (1186, 452),
            (1221, 513),
            (1291, 513),
            (1256, 574),
            (1291, 635),
            (1362, 635),
            (1397, 574),
            (1433, 635),
            (1503, 635),
            (1538, 574),
            (1503, 513),
            (1573, 513),
            (1609, 452),
            (1573, 392),
            (1503, 392),
            (1538, 331),
            (1503, 270),
            (1433, 270),
            (1397, 331),
            (1362, 270),
            (1362, 392),
            (1327, 452),
            (1362, 513),
            (1433, 513),
            (1468, 452),
            (1433, 392),
        ]
        self.point_coords = [
            (1397, 129),
            (1468, 169),
            (1538, 129),
            (1609, 169),
            (1609, 250),
            (1679, 290),
            (1679, 372),
            (1750, 412),
            (1750, 493),
            (1679, 533),
            (1679, 615),
            (1609, 655),
            (1609, 736),
            (1538, 766),
            (1468, 736),
            (1397, 776),
            (1327, 736),
            (1256, 776),
            (1186, 736),
            (1186, 655),
            (1115, 615),
            (1115, 533),
            (1045, 493),
            (1045, 412),
            (1115, 372),
            (1115, 290),
            (1186, 250),
            (1186, 169),
            (1256, 129),
            (1327, 169),
            (1327, 250),
            (1256, 290),
            (1256, 372),
            (1186, 412),
            (1186, 493),
            (1256, 533),
            (1256, 615),
            (1327, 655),
            (1397, 615),
            (1468, 655),
            (1538, 615),
            (1538, 533),
            (1609, 493),
            (1609, 412),
            (1538, 372),
            (1538, 290),
            (1468, 250),
            (1397, 290),
            (1397, 372),
            (1327, 412),
            (1327, 493),
            (1397, 533),
            (1468, 493),
            (1468, 412),
        ]


params = Params()


class CheckInitMap:
    def __init__(self) -> None:
        self.initialized = False
        self.tile_states = None
        self.port_states = None
        self.prob_states = None


checkInitMap = CheckInitMap()


def check_init_map(state):
    if not checkInitMap.initialized:
        return True

    if (checkInitMap.tile_states != state[0:114]).any():
        return True

    if (checkInitMap.prob_states != state[133:361]).any():
        return True

    if (checkInitMap.port_states != state[361:415]).any():
        return True

    return False


def draw_outlined_text(draw, text, font, pos, color, opx):
    o_color = (255 - color[0], 255 - color[1], 255 - color[2])
    x = pos[0]
    y = pos[1]
    draw.text((x + opx, y + opx), text, o_color, font)
    draw.text((x - opx, y + opx), text, o_color, font)
    draw.text((x + opx, y - opx), text, o_color, font)
    draw.text((x - opx, y - opx), text, o_color, font)
    draw.text(pos, text, color, font)


def init_map(state):
    sprites.background = sprites._background_.copy()
    checkInitMap.initialized = True
    checkInitMap.tile_states = state[0:114].copy()
    checkInitMap.prob_states = state[133:361].copy()
    checkInitMap.port_states = state[361:415].copy()

    tiles = []
    for i in range(19):
        tiles.append(np.where(state[6 * i : 6 + 6 * i] == 1)[0][0])

    for i in range(19):
        sprites.background.paste(
            sprites.tiles[tiles[i]],
            (params.tile_center_pos[i][0] - 66, params.tile_center_pos[i][1] - 76),
            sprites.tiles[tiles[i]],
        )

    probs = []
    for i in range(19):
        probs.append(np.where(state[133 + 12 * i : 145 + 12 * i] == 1)[0][0] + 1)

    for i in range(19):
        if probs[i] != 1:
            sprites.background.paste(
                sprites.probs[probs[i]],
                (params.tile_center_pos[i][0] - 27, params.tile_center_pos[i][1] - 1),
                sprites.probs[probs[i]],
            )

    ports = []
    for i in range(9):
        ports.append(np.where(state[361 + 6 * i : 361 + 6 * (i + 1)] == 1)[0][0])

    for i in range(9):
        sprites.background.paste(
            sprites.ports[ports[i]], params.port_topleft_pos[i], sprites.ports[ports[i]]
        )


def get_state_image(state=None):
    if state is None:
        return sprites.background

    if check_init_map(state):
        init_map(state)

    bg = sprites.background.copy()

    #  Draw
    draw = ImageDraw.ImageDraw(bg)

    #  Vẽ Robber
    rob_pos = np.where(state[114:133] == 1)[0][0]
    bg.paste(
        sprites.robber,
        (
            params.tile_center_pos[rob_pos][0] - 55,
            params.tile_center_pos[rob_pos][1] - 27,
        ),
        sprites.robber,
    )

    #  Tài nguyên ngân hàng, thẻ phát triển
    bank_res = state[415:421]
    for i in range(6):
        if bank_res[i] == 0:
            text = "0"
        else:
            text = "?"

        draw_outlined_text(draw, text, params.font24, (455 + 50 * i, 182), (0, 0, 0), 1)

    #  Nguyên liệu của bản thân
    my_res = state[421:426]
    for i in range(5):
        text = str(int(my_res[i]))
        draw_outlined_text(draw, text, params.font24, (455 + 50 * i, 732), (0, 0, 0), 1)

    #  Nguyên liệu còn lại trong kho
    my_inv_res = state[1268:1273]
    for i in range(5):
        text = str(int(my_inv_res[i]))
        draw_outlined_text(draw, text, params.font24, (455 + 50 * i, 702), (0, 0, 0), 1)

    #  Thẻ phát triển của bản thân
    my_dev = state[426:431]
    for i in range(5):
        text = str(int(my_dev[i]))
        draw_outlined_text(draw, text, params.font24, (455 + 50 * i, 807), (0, 0, 0), 1)

    #  Điểm của bản thân
    my_score = str(int(state[431]))
    bbox = draw.textbbox((0, 0), my_score, params.font28)
    draw_outlined_text(
        draw,
        my_score,
        params.font28,
        (350 - bbox[2] / 2, 832 - bbox[3]),
        (255, 255, 255),
        1,
    )

    #  Điểm của 3 người chơi khác
    for i in range(3):
        score = str(int(state[631 + 185 * i]))
        bbox = draw.textbbox((0, 0), score, params.font28)
        draw_outlined_text(
            draw,
            score,
            params.font28,
            (350 - bbox[2] / 2, 382 + 150 * i - bbox[3]),
            (255, 255, 255),
            1,
        )

    #  Số thẻ knight đã dùng và con đường dài nhất của bản thân
    num_knight = str(int(state[612]))
    bbox = draw.textbbox((0, 0), num_knight, params.font24)
    draw_outlined_text(
        draw, num_knight, params.font24, (310 - bbox[2] / 2, 777), (0, 0, 0), 1
    )
    longest_road = str(int(state[613]))
    bbox = draw.textbbox((0, 0), longest_road, params.font24)
    draw_outlined_text(
        draw, longest_road, params.font24, (380 - bbox[2] / 2, 777), (0, 0, 0), 1
    )

    #  Số thẻ knight đã dùng và con đường dài nhất của 3 người chơi khác
    for i in range(3):
        num_knight = str(int(state[812 + 185 * i]))
        bbox = draw.textbbox((0, 0), num_knight, params.font24)
        draw_outlined_text(
            draw,
            num_knight,
            params.font24,
            (310 - bbox[2] / 2, 327 + 150 * i),
            (0, 0, 0),
            1,
        )
        longest_road = str(int(state[813 + 185 * i]))
        bbox = draw.textbbox((0, 0), longest_road, params.font24)
        draw_outlined_text(
            draw,
            longest_road,
            params.font24,
            (380 - bbox[2] / 2, 327 + 150 * i),
            (0, 0, 0),
            1,
        )

    #  Tổng số thẻ tài nguyên và tổng số thẻ phát triển của 3 người chơi khác
    for i in range(3):
        num_res = str(int(state[629 + 185 * i]))
        draw_outlined_text(
            draw, num_res, params.font24, (455, 282 + 150 * i), (0, 0, 0), 1
        )
        num_dev = str(int(state[630 + 185 * i]))
        draw_outlined_text(
            draw, num_dev, params.font24, (455, 357 + 150 * i), (0, 0, 0), 1
        )

    #  Đường của bản thân
    my_roads = np.where(state[432:504] == 1)[0]
    draw_roads(bg, my_roads, 0)

    #  Đường của 3 người chơi khác
    for i in range(3):
        roads = np.where(state[632 + 185 * i : 704 + 185 * i] == 1)[0]
        draw_roads(bg, roads, i + 1)

    #  Nhà và thành phố của bản thân
    setts = np.where(state[504:558] == 1)[0]
    draw_setts(bg, setts, 0, city=False)
    cities = np.where(state[558:612] == 1)[0]
    draw_setts(bg, cities, 0, city=True)

    #  Nhà và thành phố của 3 người chơi khác
    for i in range(3):
        setts = np.where(state[704 + 185 * i : 758 + 185 * i])[0]
        draw_setts(bg, setts, i + 1, city=False)
        cities = np.where(state[758 + 185 * i : 812 + 185 * i])[0]
        draw_setts(bg, cities, i + 1, city=True)

    #  Phase
    phase = np.where(state[1273:1286] == 1)[0][0]
    text = f"Phase {phase}: " + phase_annotations[phase]
    draw_outlined_text(draw, text, params.font28, (280, 65), (255, 255, 255), 1)

    #  Điểm đặt thứ nhất
    try:
        pos = np.where(state[1204:1258] == 1)[0][0]
        bg.paste(
            sprites.highlight_circle,
            (params.point_coords[pos][0] - 20, params.point_coords[pos][1] - 20),
            sprites.highlight_circle,
        )
    except:
        pass

    #  Trade offer
    for i in range(5):
        text = str(int(state[1286 + i]))
        draw_outlined_text(
            draw, text, params.font24, (455 + 300 + 50 * i, 732), (0, 0, 0), 1
        )

    #  Danh hiệu quân đội mạnh nhất
    lgar = np.where(state[1184:1188] == 1)[0]
    if len(lgar) > 0:
        idx = lgar[0]
        if idx == 0:
            y = 740
        else:
            y = 140 + 150 * idx

        bg.paste(sprites.largest_army, (290, y), sprites.largest_army)

    #  Danh hiệu con đường dài nhất
    lgrd = np.where(state[1188:1192] == 1)[0]
    if len(lgrd) > 0:
        idx = lgrd[0]
        if idx == 0:
            y = 740
        else:
            y = 140 + 150 * idx

        bg.paste(sprites.longest_road, (360, y), sprites.longest_road)

    #  Tổng xx
    try:
        xx = np.where(state[1192:1203] == 1)[0][0] + 2
    except:
        xx = -1

    if xx != -1:
        if xx <= 7:
            dice_1 = 1
            dice_2 = xx - dice_1
        else:
            dice_1 = 6
            dice_2 = xx - dice_1

        bg.paste(sprites.dice[dice_1], (770, 150), sprites.dice[dice_1])
        bg.paste(sprites.dice[dice_2], (840, 150), sprites.dice[dice_2])

    return bg


phase_annotations = [
    "Place initial settlements",
    "Choose endpoints of the road",
    "Roll the dice or use development card",
    "Discard resources",
    "Move the robber",
    "Take resources at early of match",
    "Choose a module",
    "Choose a type of resource when using development card",
    "Place a settlement",
    "Place a city",
    "Choose a type of resource to trade with bank",
    "Choose a type of resource to receive from bank",
    "Take a resource from inventory",
]


def draw_setts(bg, setts, idx, city=False):
    if city:
        img = sprites.cts[idx]
    else:
        img = sprites.sts[idx]

    for sett in setts:
        bg.paste(
            img,
            (params.point_coords[sett][0] - 27, params.point_coords[sett][1] - 27),
            img,
        )


def draw_roads(bg, roads, idx):
    for road in roads:
        pos_1 = params.point_coords[_env.ROAD_POINT[road][0]]
        pos_2 = params.point_coords[_env.ROAD_POINT[road][1]]
        x = pos_1[0] - pos_2[0]
        y = pos_1[1] - pos_2[1]
        z = x * y

        if z == 0:
            img = sprites.rds[idx]
        elif z > 0:
            img = sprites.rds_60[idx]
        else:
            img = sprites.rds_120[idx]

        pos = params.road_center_pos[road]
        bg.paste(img, (int(pos[0] - img.width / 2), int(pos[1] - img.height / 2)), img)


action_annotations = [
    "Choose the point with index 0",
    "Choose the point with index 1",
    "Choose the point with index 2",
    "Choose the point with index 3",
    "Choose the point with index 4",
    "Choose the point with index 5",
    "Choose the point with index 6",
    "Choose the point with index 7",
    "Choose the point with index 8",
    "Choose the point with index 9",
    "Choose the point with index 10",
    "Choose the point with index 11",
    "Choose the point with index 12",
    "Choose the point with index 13",
    "Choose the point with index 14",
    "Choose the point with index 15",
    "Choose the point with index 16",
    "Choose the point with index 17",
    "Choose the point with index 18",
    "Choose the point with index 19",
    "Choose the point with index 20",
    "Choose the point with index 21",
    "Choose the point with index 22",
    "Choose the point with index 23",
    "Choose the point with index 24",
    "Choose the point with index 25",
    "Choose the point with index 26",
    "Choose the point with index 27",
    "Choose the point with index 28",
    "Choose the point with index 29",
    "Choose the point with index 30",
    "Choose the point with index 31",
    "Choose the point with index 32",
    "Choose the point with index 33",
    "Choose the point with index 34",
    "Choose the point with index 35",
    "Choose the point with index 36",
    "Choose the point with index 37",
    "Choose the point with index 38",
    "Choose the point with index 39",
    "Choose the point with index 40",
    "Choose the point with index 41",
    "Choose the point with index 42",
    "Choose the point with index 43",
    "Choose the point with index 44",
    "Choose the point with index 45",
    "Choose the point with index 46",
    "Choose the point with index 47",
    "Choose the point with index 48",
    "Choose the point with index 49",
    "Choose the point with index 50",
    "Choose the point with index 51",
    "Choose the point with index 52",
    "Choose the point with index 53",
    "Roll the dice",
    "Use knight",
    "Use road building",
    "Use year of plenty",
    "Use monopoly",
    "Choose lumber to receive",
    "Choose brick to receive",
    "Choose wool to receive",
    "Choose grain to receive",
    "Choose ore to receive",
    "Choose the tile with index 0",
    "Choose the tile with index 1",
    "Choose the tile with index 2",
    "Choose the tile with index 3",
    "Choose the tile with index 4",
    "Choose the tile with index 5",
    "Choose the tile with index 6",
    "Choose the tile with index 7",
    "Choose the tile with index 8",
    "Choose the tile with index 9",
    "Choose the tile with index 10",
    "Choose the tile with index 11",
    "Choose the tile with index 12",
    "Choose the tile with index 13",
    "Choose the tile with index 14",
    "Choose the tile with index 15",
    "Choose the tile with index 16",
    "Choose the tile with index 17",
    "Choose the tile with index 18",
    "Buy a road",
    "Buy a settlement",
    "Buy a city",
    "Buy a development card",
    "Create a trade offer with bank",
    "End of turn",
    "Choose lumber to give (or discard)",
    "Choose brick to give (or discard)",
    "Choose wool to give (or discard)",
    "Choose grain to give (or discard)",
    "Choose ore to give (or discard)",
    "Take a resource from inventory",
]


def get_description(action):
    if action < 0 or action >= _env.getActionSize():
        return ""

    return action_annotations[action]


class Env_components:
    def __init__(self, env, winner, list_other) -> None:
        self.env = env
        self.winner = winner
        self.list_other = list_other


def get_env_components():
    env = _env.initEnv()
    winner = _env.checkEnded(env)
    list_other = np.array([-1, 1, 2, 3])
    np.random.shuffle(list_other)

    env_components = Env_components(env, winner, list_other)
    return env_components


def get_main_player_state(
    env_components: Env_components, list_agent, list_data, action=None
):
    if not action is None:
        _env.stepEnv(env_components.env, action)

    env_components.winner = _env.checkEnded(env_components.env)
    if env_components.winner == -1:
        while env_components.env[230] < MAX_TURN_IN_ONE_GAME:
            p_idx = int(env_components.env[244])
            if env_components.list_other[p_idx] == -1:
                break

            state = _env.getAgentState(env_components.env)
            agent = list_agent[env_components.list_other[p_idx] - 1]
            data = list_data[env_components.list_other[p_idx] - 1]
            action, data = agent(state, data)
            _env.stepEnv(env_components.env, action)
            env_components.winner = _env.checkEnded(env_components.env)
            if env_components.winner != -1:
                break

        if (
            env_components.winner != -1
            or env_components.env[230] == MAX_TURN_IN_ONE_GAME
        ):
            env_components.env[280] = 1
    else:
        env_components.env[280] = 1

    if env_components.env[280] == 0:
        state = _env.getAgentState(env_components.env)
        win = -1
    else:
        env_components.env[np.array([68, 110, 152, 194])] += env_components.env[
            np.array([67, 109, 151, 193])
        ]
        my_idx = np.where(env_components.list_other == -1)[0][0]
        env = env_components.env.copy()
        env[229] = 2
        env[244] = my_idx
        state = _env.getAgentState(env)
        if my_idx == env_components.winner:
            win = 1
        else:
            win = 0

        #  Chạy turn cuối cho 3 bot hệ thống
        for p_idx in range(4):
            if p_idx != my_idx:
                env[244] = p_idx
                _state = _env.getAgentState(env)
                agent = list_agent[env_components.list_other[p_idx] - 1]
                data = list_data[env_components.list_other[p_idx] - 1]
                action, data = agent(_state, data)

    return win, state, env_components



import numpy as np
from PIL import Image, ImageDraw, ImageEnhance, ImageFont

from src.Base.Splendor_v3 import env as _env
from env import SHORT_PATH

NORMAL_CARD, NOBLE_CARD = _env.normal_cards_infor, _env.noble_cards_infor
IMG_PATH = SHORT_PATH + "src/Base/Splendor_v2/images/"

tl = 3
BG_SIZE = (np.array([2048, 2384]) / tl).astype(np.int64)
BOARD_SIZE = (int(BG_SIZE[0] * 16 / 9), BG_SIZE[1])
CARD_SIZE = (np.array([300, 425]) / (tl * 0.81)).astype(np.int64)
CARD_NOBLE_SIZE = (np.array([300, 300]) / (tl * 0.81)).astype(np.int64)
TOKEN_SIZE = (int(CARD_SIZE[0] * 0.3), int(CARD_SIZE[1] * 0.3))

action_description = {
    0: "Taget thẻ thứ 1",
    1: "Taget thẻ thứ 2",
    2: "Taget thẻ thứ 3",
    3: "Taget thẻ thứ 4",
    4: "Taget thẻ thứ 5",
    5: "Taget thẻ thứ 6",
    6: "Taget thẻ thứ 7",
    7: "Taget thẻ thứ 8",
    8: "Taget thẻ thứ 9",
    9: "Taget thẻ thứ 10",
    10: "Taget thẻ thứ 11",
    11: "Taget thẻ thứ 12",
    12: "Taget thẻ đang úp thứ 1",
    13: "Taget thẻ đang úp thứ 2",
    14: "Taget thẻ đang úp thứ 3",
}


class Env_components:
    def __init__(self, env, winner, list_other, lv1, lv2, lv3, list_action) -> None:
        self.env = env
        self.winner = winner
        self.list_other = list_other
        self.lv1 = lv1
        self.lv2 = lv2
        self.lv3 = lv3
        self.cc = 0
        self.list_action = list_action


def get_description(action):
    if action < 0 or action >= _env.getActionSize():
        return ""
    return f"{action_description[action]}"


def get_env_components():
    env, lv1, lv2, lv3 = _env.initEnv()
    winner = _env.checkEnded(env)
    list_other = np.array([-1, 1, 2, 3])
    np.random.shuffle(list_other)
    while list_other[-1] == -1:
        np.random.shuffle(list_other)
    state = _env.getAgentState(env, lv1, lv2, lv3)
    list_action = _env.getValidActions(state)
    env_components = Env_components(env, winner, list_other, lv1, lv2, lv3, list_action)
    return env_components


class Sprites:
    def __init__(self) -> None:
        self.font = ImageFont.truetype(
            "src/ImageFonts/FreeMonoBoldOblique.ttf", int(120 / tl)
        )
        self.font_max = ImageFont.truetype(
            "src/ImageFonts/FreeMonoBoldOblique.ttf", int(200 / tl)
        )
        self.font2 = ImageFont.truetype(
            "src/ImageFonts/FreeMonoBoldOblique.ttf", int(60 / tl)
        )

        self.im = Image.new("RGB", BOARD_SIZE, "black")
        self.background = Image.open(IMG_PATH + "bg.webp").resize(BG_SIZE)
        self.list_token_name = ["red", "blue", "green", "black", "white", "yellow"]

        self.list_img_card = [
            Image.open(IMG_PATH + f"Cards/{card}.png").resize(CARD_SIZE)
            for card in range(90)
        ]
        self.list_img_card_noble = [
            Image.open(IMG_PATH + f"Cards/{card}.png").resize(CARD_NOBLE_SIZE)
            for card in range(90, 100)
        ]
        self.list_img_card_hide = [
            Image.open(IMG_PATH + f"Cards/hide_card_{card}.png").resize(CARD_SIZE)
            for card in range(1, 4)
        ]
        self.list_img_token = [
            Image.open(IMG_PATH + f"Tokens/{token}.png").resize(TOKEN_SIZE)
            for token in self.list_token_name
        ]


class Params:
    def __init__(self) -> None:
        pass


_d_ = int(BG_SIZE[0] * 0.02)
_d2_ = int(BG_SIZE[0] * 0.05)
params = Params()
sprites = Sprites()


def draw_cards_image(
    background,
    list_id_noble=np.full(5, 0),
    list_id_normal=np.full((4, 3), 0),
    list_hide_card=np.full(3, 0),
):
    #  Draw cards image on background image
    y = int(BG_SIZE[1] * 0.77)
    for d_y in range(4):
        x = int(BG_SIZE[0] * 0.01)
        for d_x in range(5):
            if d_y == 3:
                if d_x < len(list_id_noble):
                    background.paste(
                        sprites.list_img_card_noble[list_id_noble[d_x]],
                        (x, int(y * 2.1)),
                    )
            else:
                if d_x == 0:
                    if list_hide_card[d_y] == 1:
                        background.paste(sprites.list_img_card_hide[d_y], (x, y))
                else:
                    id_card_in_list = 4 * d_y + d_x - 1
                    if list_id_normal[id_card_in_list] != -1:
                        background.paste(
                            sprites.list_img_card[list_id_normal[id_card_in_list]],
                            (x, y),
                        )
            x += CARD_SIZE[0] + _d_
        y -= CARD_SIZE[1] + _d_


def draw_tokens_image(
    im,
    list_token_board=np.full(6, 1),
    list_token_const=np.full(5, 1),
    list_token=np.full(6, 1),
    list_token_taken=np.full(5, 1),
):
    #  Draw tokens on board
    x = int(BG_SIZE[0] * 0.22)
    y = int(BOARD_SIZE[1] * 0.04)
    for i in range(6):
        im.paste(sprites.list_img_token[i], (x, y))
        ImageDraw.Draw(im).text(
            (x + TOKEN_SIZE[0], y + int(TOKEN_SIZE[1] / 3)),
            f"{list_token_board[i]}",
            fill="white",
            font=sprites.font,
        )
        x += TOKEN_SIZE[0] + _d2_

    #  Draw tokens const
    y = int(BOARD_SIZE[1] * 0.07)
    y_ = int(BOARD_SIZE[1] * 0.2)
    x = int(BG_SIZE[0] * 1.02)
    ImageDraw.Draw(im).text(
        (x, int(BOARD_SIZE[1] * 0.03)),
        f"Nguyên liệu mặc định:",
        fill="white",
        font=sprites.font2,
    )
    for i in range(5):
        im.paste(sprites.list_img_token[i], (x, y))
        ImageDraw.Draw(im).text(
            (x + TOKEN_SIZE[0], y + int(TOKEN_SIZE[1] / 3)),
            f"{list_token_const[i]}",
            fill="white",
            font=sprites.font,
        )
        x += TOKEN_SIZE[0] + _d2_

    #  Draw tokens on hand
    x = int(BG_SIZE[0] * 1.02)
    ImageDraw.Draw(im).text(
        (x, int(BOARD_SIZE[1] * 0.23)),
        f"Nguyên liệu của bản thân:",
        fill="white",
        font=sprites.font2,
    )
    for i in range(6):
        im.paste(sprites.list_img_token[i], (x, y + y_))
        ImageDraw.Draw(im).text(
            (x + TOKEN_SIZE[0], y + y_ + int(TOKEN_SIZE[1] / 3)),
            f"{list_token[i]}",
            fill="white",
            font=sprites.font,
        )
        x += TOKEN_SIZE[0] + _d2_

    #  Draw tokens taken
    #  x = int(BG_SIZE[0]*1.02)
    #  y = int(BG_SIZE[1]*0.8)
    #  ImageDraw.Draw(im).text((x, y*0.95), f'Nguyên liệu đã lấy trong lượt:', fill= 'white', font = sprites.font2)
    #  for i in range(5):
    #      im.paste(sprites.list_img_token[i], (x, y  ))
    #      ImageDraw.Draw(im).text((x + TOKEN_SIZE[0], y + int(TOKEN_SIZE[1]/3)), f'{list_token_taken[i]}', fill= 'white', font = sprites.font)
    #      x += TOKEN_SIZE[0] + _d2_


def draw_down_card(im, state):
    down_card = state[175:208].reshape(3, 11)
    list_down_card = []
    for card in down_card:
        for i in range(len(NORMAL_CARD)):
            if (card == NORMAL_CARD[i]).all():
                list_down_card.append(i)

    x = int(BOARD_SIZE[0] * 0.58)
    y = int(BOARD_SIZE[1] * 0.43)
    for i in range(len(list_down_card)):
        im.paste(sprites.list_img_card[list_down_card[i]], (x, y))
        x += CARD_SIZE[0] + _d_


def draw_score(im, state):
    scores = [state[17], state[213], state[214], state[215]]
    x = int(im.size[0] * 0.925)
    y = int(BOARD_SIZE[1] * 0.08)
    for agent in range(4):
        ImageDraw.Draw(im).text(
            (x, y - int(BOARD_SIZE[1] * 0.03)),
            f"Điểm:",
            fill="white",
            font=sprites.font2,
        )
        ImageDraw.Draw(im).text(
            (x, y), f"{scores[agent]}", fill="white", font=sprites.font_max
        )
        if agent == 0:
            ImageDraw.Draw(im).text(
                (x, y + int(BOARD_SIZE[1] * 0.11)),
                f"YOU",
                fill="white",
                font=sprites.font,
            )
        y += int(im.size[1] / 4)


def draw_line(im, color="White", width=3):
    w = int(im.size[0] * 0.92)
    for i in range(3):
        h = int(im.size[1] / 4) * (i + 1)
        shape = [(w, h), (im.size[0], h)]
        ImageDraw.Draw(im).line(shape, fill=color, width=width)

    ImageDraw.Draw(im).line([(w, 0), (w, im.size[1])], fill=color, width=width)


def get_id_card(background, state=None):
    cards_noble = state[150:175].reshape(5, 5)
    list_id_noble = []
    for card in cards_noble:
        for i in range(len(NOBLE_CARD)):
            if (card == NOBLE_CARD[i]).all():
                list_id_noble.append(i)

    card_normal = state[18:150].reshape(12, 11)
    list_id_normal = []
    for card in card_normal:
        for i in range(len(NORMAL_CARD)):
            if (card == NORMAL_CARD[i]).all():
                list_id_normal.append(i)
    list_card_normal_convert = np.full(12, -1)
    j = 0
    for i in range(len(list_id_normal)):
        if list_id_normal[i] < 40:
            list_card_normal_convert[j] = list_id_normal[i]
            j += 1
        elif list_id_normal[i] < 70:
            if j < 4:
                j = 4
            list_card_normal_convert[j] = list_id_normal[i]
            j += 1
        elif list_id_normal[i] < 90:
            if j < 8:
                j = 8
            list_card_normal_convert[j] = list_id_normal[i]
            j += 1

    list_hide_card = state[216:219]
    draw_cards_image(
        background,
        list_id_noble=list_id_noble,
        list_id_normal=list_card_normal_convert,
        list_hide_card=list_hide_card,
    )


def get_state_image(state=None):
    state = state.astype(np.int64)
    background = sprites.background.copy()
    im = sprites.im.copy()
    draw_line(im)

    get_id_card(background, state)
    im.paste(background, (0, 0))

    draw_tokens_image(
        im,
        list_token_board=state[0:6],
        list_token=state[6:12],
        list_token_const=state[12:17],
        list_token_taken=state[208:213],
    )

    draw_down_card(im, state)
    draw_score(im, state)
    return im


def get_main_player_state(
    env_components: Env_components, list_agent, list_data, action=None
):
    if not action is None:
        (
            env_components.env,
            env_components.lv1,
            env_components.lv2,
            env_components.lv3,
        ) = _env.stepEnv(
            action,
            env_components.env,
            env_components.lv1,
            env_components.lv2,
            env_components.lv3,
            env_components.list_action,
        )

    if env_components.winner == 0:  #  no ended
        while env_components.env[100] <= 400 and env_components.cc < 10000:
            idx = env_components.env[100] % 4
            state = _env.getAgentState(
                env_components.env,
                env_components.lv1,
                env_components.lv2,
                env_components.lv3,
            )
            env_components.list_action = _env.getValidActions(state)
            if env_components.list_other[idx] == -1:
                break

            agent = list_agent[env_components.list_other[idx] - 1]
            data = list_data[env_components.list_other[idx] - 1]
            action, data = agent(state, data)
            (
                env_components.env,
                env_components.lv1,
                env_components.lv2,
                env_components.lv3,
            ) = _env.stepEnv(
                action,
                env_components.env,
                env_components.lv1,
                env_components.lv2,
                env_components.lv3,
                env_components.list_action,
            )
            env_components.winner = _env.checkEnded(env_components.env)
            if env_components.winner != 0:
                break
            env_components.cc += 1

    if env_components.winner == 0:  # no winner
        state = _env.getAgentState(
            env_components.env,
            env_components.lv1,
            env_components.lv2,
            env_components.lv3,
        )
        win = -1
    else:  # have winner
        my_idx = np.where(env_components.list_other == -1)[0][0]
        env = env_components.env.copy()
        env[100] = my_idx
        state = _env.getAgentState(
            env, env_components.lv1, env_components.lv2, env_components.lv3
        )
        state[220] = 1
        if my_idx == (env_components.winner - 1):
            win = 1
        else:
            win = 0

        #  Chạy turn cuối cho 3 bot hệ thống
        for p_idx in range(4):
            if p_idx != my_idx:
                env[100] = p_idx
                _state = _env.getAgentState(
                    env, env_components.lv1, env_components.lv2, env_components.lv3
                )
                _state[220] = 1
                agent = list_agent[env_components.list_other[p_idx] - 1]
                data = list_data[env_components.list_other[p_idx] - 1]
                action, data = agent(_state, data)

    return win, state, env_components



import numpy as np
from PIL import Image, ImageDraw, ImageFont

from src.Base.Imploding_Kitten import env as _env
from env import SHORT_PATH

IMG_PATH = SHORT_PATH + "src/Base/Exploding_Kitten/images/"


class Params:
    def __init__(self) -> None:
        self.card_order = [
            "Nope",
            "Attack",
            "Skip",
            "Favor",
            "Shuffle",
            "Future",
            "Taco",
            "Rainbow",
            "Beard",
            "Potato",
            "Melon",
            "Defuse",
            "Exploding",
        ]
        self.im_card_order = [
            "Reverse",
            "DrawFromBottom",
            "Feral",
            "AlterTheFuture",
            "TargetAttack",
            "Imploding",
        ]
        self.all_card_order = [
            "Nope",
            "Attack",
            "Skip",
            "Favor",
            "Shuffle",
            "Future",
            "Taco",
            "Rainbow",
            "Beard",
            "Potato",
            "Melon",
            "Reverse",
            "DrawFromBottom",
            "Feral",
            "AlterTheFuture",
            "TargetAttack",
            "Defuse",
        ]
        self.font32 = ImageFont.FreeTypeFont("src/ImageFonts/arial.ttf", 32)
        self.font24 = ImageFont.FreeTypeFont("src/ImageFonts/arial.ttf", 24)


params = Params()


class Sprites:
    def __init__(self) -> None:
        self.background = Image.open(IMG_PATH + "im_background.png").resize((2100, 900))
        self._background_ = self.background.copy()
        self.card_back = (
            Image.open(IMG_PATH + "Cardback.png").resize((120, 168)).convert("RGBA")
        )
        self.cards = []
        self.sm_cards = []
        for name in params.card_order:
            self.cards.append(
                Image.open(IMG_PATH + f"{name}.png").resize((120, 168)).convert("RGBA")
            )
            self.sm_cards.append(
                Image.open(IMG_PATH + f"{name}.png").resize((100, 140)).convert("RGBA")
            )

        self.im_cards = []
        self.sm_im_cards = []
        for name in params.im_card_order:
            self.im_cards.append(
                Image.open(IMG_PATH + f"{name}.png").resize((120, 168)).convert("RGBA")
            )
            self.sm_im_cards.append(
                Image.open(IMG_PATH + f"{name}.png").resize((100, 140)).convert("RGBA")
            )


sprites = Sprites()


def draw_outlined_text(draw, text, font, pos, color, opx):
    o_color = (255 - color[0], 255 - color[1], 255 - color[2])
    x = pos[0]
    y = pos[1]
    draw.text((x + opx, y + opx), text, o_color, font)
    draw.text((x - opx, y + opx), text, o_color, font)
    draw.text((x + opx, y - opx), text, o_color, font)
    draw.text((x - opx, y - opx), text, o_color, font)
    draw.text(pos, text, color, font)


def get_state_image(state=None):
    if state is None:
        return sprites.background

    bg = sprites.background.copy()

    #  Draw
    draw = ImageDraw.ImageDraw(bg)

    #  Hold in hand
    temp = state[0:17].astype(int)
    for i in range(17):
        text = str(temp[i])
        draw_outlined_text(
            draw, text, params.font32, (68 + 110 * i, 827), (255, 255, 255), 1
        )

    #  Discard pile
    temp = state[17:35].astype(int)
    for i in range(18):
        text = str(temp[i])
        bbox = draw.textbbox((0, 0), text, params.font32)
        draw_outlined_text(
            draw,
            text,
            params.font32,
            (162 + 110 * i - bbox[2], 720),
            (255, 255, 255),
            1,
        )

    #  Draw pile
    text = str(int(state[35]))
    bbox = draw.textbbox((0, 0), text, params.font32)
    draw_outlined_text(
        draw, text, params.font32, (1050 - bbox[2] // 2, 380), (255, 255, 255), 1
    )

    #  Have to draw
    text = str(int(state[100]))
    bbox = draw.textbbox((0, 0), text, params.font32)
    draw_outlined_text(
        draw, text, params.font32, (1050 - bbox[2] // 2, 420), (0, 0, 0), 1
    )

    #  Others
    for i in range(5):
        if state[116 + i] == 0:
            text = "Exploded"
            font = params.font24
            b_h = 5
        else:
            text = str(int(state[122 + i]))
            font = params.font32
            b_h = 0

        bbox = draw.textbbox((0, 0), text, font)
        draw_outlined_text(
            draw,
            text,
            font,
            (110 + 470 * i - bbox[2] // 2, 160 + b_h),
            (255, 255, 255),
            1,
        )

    #  Last action
    a = np.where(state[101:116] == 1)[0]
    if len(a) == 1:
        last_action = a[0]
        if last_action < 6:
            bg.paste(sprites.cards[last_action], (300, 250))
        elif last_action == 11:
            bg.paste(sprites.im_cards[0], (300, 250))
        elif last_action == 12:
            bg.paste(sprites.im_cards[1], (300, 250))
        elif last_action == 13:
            bg.paste(sprites.im_cards[3], (300, 250))
        elif last_action == 14:
            bg.paste(sprites.im_cards[4], (300, 250))
        else:
            if last_action == 6:
                text = "Draw card"
            elif last_action == 7:
                text = "Double"
            elif last_action == 8:
                text = "Triple"
            elif last_action == 9:
                text = "5 kinds of cards"
            else:
                text = ""

            bbox = draw.textbbox((0, 0), text, params.font32)
            draw_outlined_text(
                draw, text, params.font32, (360 - bbox[2] // 2, 300), (0, 0, 0), 1
            )

    #  Exploded
    if state[121] == 0:
        text = "Exploded"
        bbox = draw.textbbox((0, 0), text, params.font32)
        draw_outlined_text(
            draw, text, params.font32, (1050 - bbox[2] // 2, 630), (0, 0, 0), 1
        )

    #  See the future
    temp = []
    for i in range(3):
        try:
            temp.append(np.where(state[38 + 19 * i : 57 + 19 * i] == 1)[0][0])
        except:
            pass

    w_ = len(temp) * 140 - 20
    s_ = 1520 - w_ // 2
    for i in range(len(temp)):
        idx = temp[i]
        if idx < 11:
            card = sprites.cards[idx]
        elif idx < 16:
            card = sprites.im_cards[idx - 11]
        elif idx == 16:
            card = sprites.cards[-2]
        elif idx == 17:
            card = sprites.cards[-1]
        else:
            card = sprites.im_cards[-1]

        bg.paste(card, (s_ + 140 * i, 250))

    #  Nope
    if state[37] == 1:
        text = "Nope"
        bbox = draw.textbbox((0, 0), text, params.font32)
        draw_outlined_text(
            draw, text, params.font32, (360 - bbox[2] // 2, 420), (0, 0, 0), 1
        )

    #  Cards have been discarded
    temp = []
    for i in range(16):
        for k in range(int(state[128 + i])):
            temp.append(i)

    w_ = len(temp) * 110 - 10
    s_ = 1050 - w_ // 2
    for i in range(len(temp)):
        idx = temp[i]
        if idx < 11:
            card = sprites.sm_cards[idx]
        elif idx < 16:
            card = sprites.sm_im_cards[idx - 11]

        bg.paste(card, (s_ + 110 * i, 470))

    #  Imploded
    if state[127] == 1:
        bg.paste(sprites.im_cards[-1], (680, 250))

    return bg


temp = [
    "Nope",
    "Attack",
    "Skip",
    "Favor",
    "Shuffle",
    "Future",
    "Taco",
    "Rainbow",
    "Beard",
    "Potato",
    "Melon",
    "Reverse",
    "DrawFromBottom",
    "Feral",
    "AlterTheFuture",
    "TargetAttack",
    "Defuse",
]
action_annotations = (
    [
        "Nope",
        "Attack",
        "Skip",
        "Favor",
        "Shuffle",
        "See the future",
        "Draw card",
        "Play doublecard",
        "Play triplecard",
        "Play five kinds of card",
        "Not Nope",
        "Reverse",
        "Draw from bottom",
        "Alter the future",
        "Targeted Attack",
    ]
    + [f"Choose player {i+1} to rob" for i in range(5)]
    + [f'Choose "{temp[i]}"-card to give' for i in range(17)]
    + [f'Ask about "{temp[i]}"-card' for i in range(17)]
    + [f'Choose "{temp[i]}"-card from Discard pile' for i in range(17)]
    + [f"Choose the {i+1}-th combination to Alter" for i in range(6)]
    + [f'Discard "{temp[i]}"-card' for i in range(16)]
)


def get_description(action):
    if action < 0 or action >= _env.getActionSize():
        return ""

    return action_annotations[action]


class Env_components:
    def __init__(self, env, draw, discard, winner, list_other, turn) -> None:
        self.env = env
        self.draw = draw
        self.discard = discard
        self.winner = winner
        self.list_other = list_other
        self.turn = turn


def get_env_components():
    env, draw, discard = _env.initEnv()
    winner = _env.checkEnded(env)
    list_other = np.array([-1, 1, 2, 3, 4, 5])
    np.random.shuffle(list_other)
    turn = 0
    return Env_components(env, draw, discard, winner, list_other, turn)


def get_main_player_state(
    env_components: Env_components, list_agent, list_data, action=None
):
    if not action is None:
        env_components.env, env_components.draw, env_components.discard = _env.stepEnv(
            env_components.env, env_components.draw, env_components.discard, action
        )
        env_components.turn += 1

    env_components.winner = _env.checkEnded(env_components.env)
    while env_components.winner == -1 and env_components.turn < 300:
        phase = env_components.env[89]
        main_id = env_components.env[77]
        nope_id = env_components.env[95]
        last_action = env_components.env[94]
        if phase == 0:
            pIdx = int(main_id)
        elif phase == 1:
            pIdx = int(nope_id)
        elif phase == 2:
            pIdx = int(main_id)
        elif phase == 3:
            if last_action == 3:
                pIdx = int(env_components.env[96])
            else:
                pIdx = int(main_id)
        elif phase == 4:
            pIdx = int(main_id)
        elif phase == 5:
            pIdx = int(main_id)

        if env_components.list_other[pIdx] == -1:
            break

        state = _env.getAgentState(
            env_components.env, env_components.draw, env_components.discard
        )
        agent = list_agent[env_components.list_other[pIdx] - 1]
        data = list_data[env_components.list_other[pIdx] - 1]
        action, data = agent(state, data)
        env_components.env, env_components.draw, env_components.discard = _env.stepEnv(
            env_components.env, env_components.draw, env_components.discard, action
        )
        env_components.turn += 1
        env_components.winner = _env.checkEnded(env_components.env)

    if env_components.winner == -1 and env_components.turn < 300:
        state = _env.getAgentState(
            env_components.env, env_components.draw, env_components.discard
        )
        win = -1
    else:
        env = env_components.env.copy()
        env[89] = 0
        my_idx = np.where(env_components.list_other == -1)[0][0]
        env[77] = my_idx
        env[78:83] = _env.nopeTurn(my_idx)
        state = _env.getAgentState(env, env_components.draw, env_components.discard)
        if _env.checkEnded(env) == my_idx:
            win = 1
        else:
            win = 0

        for pIdx in range(5):
            env[77] = pIdx
            env[78:83] = _env.nopeTurn(pIdx)
            if pIdx != my_idx:
                _state = _env.getAgentState(
                    env, env_components.draw, env_components.discard
                )
                agent = list_agent[env_components.list_other[pIdx] - 1]
                data = list_data[env_components.list_other[pIdx] - 1]
                action, data = agent(_state, data)

    return win, state, env_components



import numpy as np
from numba import njit

#############################

__NORMAL_CARD__ = np.array(
    [
        [0, 0, 0, 1, 0, 0, 2, 2, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 3, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 1, 1, 0, 2, 1],
        [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 2],
        [0, 0, 0, 1, 0, 0, 0, 3, 1, 0, 1],
        [0, 0, 0, 1, 0, 0, 1, 1, 0, 1, 1],
        [1, 0, 0, 1, 0, 0, 0, 0, 0, 4, 0],
        [0, 0, 0, 1, 0, 0, 2, 1, 0, 2, 0],
        [0, 0, 1, 0, 0, 0, 2, 0, 2, 0, 1],
        [0, 0, 1, 0, 0, 0, 0, 0, 2, 2, 0],
        [0, 0, 1, 0, 0, 0, 1, 0, 1, 1, 1],
        [0, 0, 1, 0, 0, 0, 2, 0, 1, 1, 1],
        [0, 0, 1, 0, 0, 0, 1, 1, 3, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0, 0, 2, 1],
        [0, 0, 1, 0, 0, 0, 0, 0, 0, 3, 0],
        [1, 0, 1, 0, 0, 0, 4, 0, 0, 0, 0],
        [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 4],
        [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 3],
        [0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 2],
        [0, 1, 0, 0, 0, 0, 0, 0, 1, 2, 2],
        [0, 1, 0, 0, 0, 0, 1, 0, 0, 3, 1],
        [0, 1, 0, 0, 0, 0, 2, 0, 0, 0, 2],
        [0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1],
        [0, 1, 0, 0, 0, 0, 0, 2, 1, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 2, 2, 1, 0],
        [0, 0, 0, 0, 0, 1, 1, 1, 2, 1, 0],
        [0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 3],
        [1, 0, 0, 0, 0, 1, 0, 0, 4, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 2, 0, 2, 0],
        [0, 0, 0, 0, 0, 1, 2, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0],
        [0, 0, 0, 0, 0, 1, 0, 3, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 1, 0, 2, 0, 0],
        [0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1],
        [1, 0, 0, 0, 1, 0, 0, 4, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 1, 2, 0, 0, 2],
        [0, 0, 0, 0, 1, 0, 0, 0, 3, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0, 2, 0, 2],
        [0, 0, 0, 0, 1, 0, 3, 0, 1, 1, 0],
        [0, 0, 0, 0, 1, 0, 1, 2, 1, 0, 1],
        [1, 0, 0, 1, 0, 0, 0, 3, 0, 2, 2],
        [2, 0, 0, 1, 0, 0, 0, 2, 0, 1, 4],
        [1, 0, 0, 1, 0, 0, 3, 0, 2, 0, 3],
        [2, 0, 0, 1, 0, 0, 0, 5, 3, 0, 0],
        [2, 0, 0, 1, 0, 0, 0, 0, 5, 0, 0],
        [3, 0, 0, 1, 0, 0, 0, 0, 6, 0, 0],
        [3, 0, 1, 0, 0, 0, 0, 6, 0, 0, 0],
        [2, 0, 1, 0, 0, 0, 1, 0, 0, 4, 2],
        [2, 0, 1, 0, 0, 0, 0, 5, 0, 0, 0],
        [2, 0, 1, 0, 0, 0, 0, 3, 0, 0, 5],
        [1, 0, 1, 0, 0, 0, 0, 2, 3, 3, 0],
        [1, 0, 1, 0, 0, 0, 3, 2, 2, 0, 0],
        [3, 1, 0, 0, 0, 0, 6, 0, 0, 0, 0],
        [2, 1, 0, 0, 0, 0, 0, 0, 0, 5, 3],
        [2, 1, 0, 0, 0, 0, 0, 0, 0, 5, 0],
        [2, 1, 0, 0, 0, 0, 0, 4, 2, 0, 1],
        [1, 1, 0, 0, 0, 0, 2, 3, 0, 3, 0],
        [1, 1, 0, 0, 0, 0, 2, 0, 0, 3, 2],
        [3, 0, 0, 0, 0, 1, 0, 0, 0, 0, 6],
        [2, 0, 0, 0, 0, 1, 5, 0, 0, 3, 0],
        [2, 0, 0, 0, 0, 1, 5, 0, 0, 0, 0],
        [1, 0, 0, 0, 0, 1, 3, 3, 0, 0, 2],
        [1, 0, 0, 0, 0, 1, 2, 0, 3, 2, 0],
        [2, 0, 0, 0, 0, 1, 4, 0, 1, 2, 0],
        [1, 0, 0, 0, 1, 0, 0, 2, 2, 0, 3],
        [1, 0, 0, 0, 1, 0, 0, 0, 3, 2, 3],
        [2, 0, 0, 0, 1, 0, 2, 1, 4, 0, 0],
        [2, 0, 0, 0, 1, 0, 3, 0, 5, 0, 0],
        [2, 0, 0, 0, 1, 0, 0, 0, 0, 0, 5],
        [3, 0, 0, 0, 1, 0, 0, 0, 0, 6, 0],
        [4, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0],
        [4, 0, 0, 1, 0, 0, 0, 6, 3, 0, 3],
        [5, 0, 0, 1, 0, 0, 0, 7, 3, 0, 0],
        [3, 0, 0, 1, 0, 0, 3, 3, 0, 3, 5],
        [3, 0, 1, 0, 0, 0, 3, 0, 3, 5, 3],
        [4, 0, 1, 0, 0, 0, 0, 0, 0, 0, 7],
        [5, 0, 1, 0, 0, 0, 0, 3, 0, 0, 7],
        [4, 0, 1, 0, 0, 0, 0, 3, 0, 3, 6],
        [3, 1, 0, 0, 0, 0, 0, 5, 3, 3, 3],
        [4, 1, 0, 0, 0, 0, 0, 0, 7, 0, 0],
        [5, 1, 0, 0, 0, 0, 3, 0, 7, 0, 0],
        [4, 1, 0, 0, 0, 0, 3, 3, 6, 0, 0],
        [5, 0, 0, 0, 0, 1, 0, 0, 0, 7, 3],
        [3, 0, 0, 0, 0, 1, 5, 3, 3, 3, 0],
        [4, 0, 0, 0, 0, 1, 0, 0, 0, 7, 0],
        [4, 0, 0, 0, 0, 1, 3, 0, 0, 6, 3],
        [3, 0, 0, 0, 1, 0, 3, 3, 5, 0, 3],
        [5, 0, 0, 0, 1, 0, 7, 0, 0, 3, 0],
        [4, 0, 0, 0, 1, 0, 6, 0, 3, 3, 0],
        [4, 0, 0, 0, 1, 0, 7, 0, 0, 0, 0],
    ],
    dtype=np.int64,
)
__NOBLE_CARD__ = np.array(
    [
        [3, 0, 4, 4, 0, 0],
        [3, 3, 0, 3, 3, 0],
        [3, 3, 3, 3, 0, 0],
        [3, 3, 0, 0, 3, 3],
        [3, 0, 3, 0, 3, 3],
        [3, 4, 0, 4, 0, 0],
        [3, 4, 0, 0, 4, 0],
        [3, 0, 3, 3, 0, 3],
        [3, 0, 4, 0, 0, 4],
        [3, 0, 0, 0, 4, 4],
    ],
    dtype=np.int64,
)


__ENV_SIZE__ = 90


@njit()
def initEnv():
    lv1 = np.arange(41, dtype=np.int64)
    lv2 = np.arange(40, 71, dtype=np.int64)
    lv3 = np.arange(70, 91, dtype=np.int64)

    np.random.shuffle(lv1[:-1])
    np.random.shuffle(lv2[:-1])
    np.random.shuffle(lv3[:-1])

    lv1[-1] = 4
    lv2[-1] = 4
    lv3[-1] = 4

    env = np.full(__ENV_SIZE__, 0, dtype=np.int64)

    env[0:6] = np.array([7, 7, 7, 7, 7, 5])

    noble_ = np.arange(10)
    np.random.shuffle(noble_)
    env[6:11] = noble_[:5]

    env[11:15] = lv1[:4]
    env[15:19] = lv2[:4]
    env[19:23] = lv3[:4]

    #  23:38:53:68:83
    for pIdx in range(4):
        temp_ = 15 * pIdx
        #  env[23+temp_:35+temp_] = 0
        env[35 + temp_ : 38 + temp_] = -1

    #  env[83] == 0 #  Turn
    #  env[84:89] = 0 #  Dùng khi lấy nguyên liệu
    #  env[89] = 0 #  1 khi game kết thúc

    return env, lv1, lv2, lv3


__STATE_SIZE__ = 267


@njit()
def getAgentState(env, lv1, lv2, lv3):
    state = np.zeros(__STATE_SIZE__, dtype=np.float64)

    state[0:6] = env[0:6]

    #  6:12:18:24:30:36 #  Thẻ Noble
    for i in range(5):
        nobleId = env[6 + i]
        if nobleId != -1:
            temp_ = 6 * i
            state[6 + temp_ : 12 + temp_] = __NOBLE_CARD__[nobleId]

    #  36:47:58:69:80:91:102:113:124:135:146:157:168 #  Thẻ normal
    for i in range(12):
        cardId = env[11 + i]
        if cardId != -1:
            temp_ = 11 * i
            state[36 + temp_ : 47 + temp_] = __NORMAL_CARD__[cardId]

    pIdx = env[83] % 4
    for i in range(4):
        pEnvIdx = (pIdx + i) % 4
        temp1 = 12 * i
        temp2 = 15 * pEnvIdx

        #  201:213:225:237:249 #  Player infor
        state[201 + temp1 : 213 + temp1] = env[23 + temp2 : 35 + temp2]

        if i == 0:
            #  168:179:190:201 #  Thẻ úp
            for j in range(3):
                cardId = env[35 + temp2 + j]
                if cardId != -1:
                    temp_ = 11 * j
                    state[168 + temp_ : 179 + temp_] = __NORMAL_CARD__[cardId]

        else:
            #  249:252:255:258 #  Đếm cấp thẻ úp
            temp_ = 3 * (i - 1)
            for j in range(3):
                cardId = env[35 + temp2 + j]
                if cardId != -1:
                    if cardId < 40:
                        state[249 + temp_] += 1
                    elif cardId < 70:
                        state[250 + temp_] += 1
                    else:
                        state[251 + temp_] += 1

    #  [258:263] #  Nguyên liệu đã lấy
    state[258:263] = env[84:89]

    #  [263]
    state[263] = env[89]

    if lv1[-1] < 40:  #  Còn thẻ ẩn cấp 1
        state[264] = 1
    if lv2[-1] < 30:  #  Còn thẻ ẩn cấp 2
        state[265] = 1
    if lv3[-1] < 20:  #  Còn thẻ ẩn cấp 3
        state[266] = 1

    return state


@njit()
def checkBuyCard(gems, perGems, price):
    temp_ = gems[0:5] + perGems
    if np.sum((price > temp_) * (price - temp_)) <= gems[5]:
        return True

    return False


@njit()
def get_card_index(arr):
    for i in range(90):
        if (__NORMAL_CARD__[i] == arr).all():
            return i

    return -1


__ACTION_SIZE__ = 194


@njit()
def getValidActions(state):
    validActions = np.zeros(__ACTION_SIZE__, dtype=np.float64)
    boardStocks = state[0:6]

    takenStocks = state[258:263]
    if (takenStocks > 0).any():  #  Đang lấy nguyên liệu
        temp_ = np.where(boardStocks[0:5] > 0)[0]
        validActions[temp_] = 1

        s_ = np.sum(takenStocks)
        if s_ == 1:
            t_ = np.where(takenStocks == 1)[0]
            if t_.shape[0] > 0:
                t_ = t_[0]
                if boardStocks[t_] < 3:
                    validActions[t_] = 0
        else:
            t_ = np.where(takenStocks == 1)[0]
            validActions[t_] = 0

        return validActions

    if np.sum(state[201:207]) > 10:  #  Thừa nguyên liệu, cần trả nguyên liệu
        temp_ = np.where(state[201:206] > 0)[0] + 185
        validActions[temp_] = 1
        return validActions

    #  Lấy nguyên liệu
    temp_ = np.where(boardStocks[0:5] > 0)[0]
    validActions[temp_] = 1

    checkReserveCard = False
    for i in range(3):
        temp_ = 11 * i
        if (state[174 + temp_ : 179 + temp_] == 0).all():
            checkReserveCard = True
            break

    #  Các action mua thẻ (và úp thẻ)
    for i in range(15):
        temp_ = 11 * i
        cardPrice = state[42 + temp_ : 47 + temp_]
        if (cardPrice > 0).any():
            card_id = get_card_index(state[36 + temp_ : 47 + temp_])
            if card_id != -1:
                if checkReserveCard and i < 12:
                    validActions[95 + card_id] = 1

                if checkBuyCard(state[201:207], state[207:212], cardPrice):
                    validActions[5 + card_id] = 1

    #  Check úp thẻ ẩn
    if checkReserveCard:
        for i in range(3):
            if state[264 + i] == 1:
                validActions[190 + i] = 1

    #  Check nếu không có action nào có thể thực hiện (bị kẹt) thì cho action bỏ lượt
    if (validActions > 0).any():
        return validActions

    validActions[193] = 1
    return validActions


@njit()
def bot_lv0(state, perData):
    validActions = getValidActions(state)
    arr_action = np.where(validActions == 1)[0]
    idx = np.random.randint(0, arr_action.shape[0])
    return arr_action[idx], perData


@njit()
def checkEnded(env):
    scoreArr = env[np.array([34, 49, 64, 79])]
    maxScore = np.max(scoreArr)
    if maxScore >= 15 and env[83] % 4 == 0:
        maxScorePlayers = np.where(scoreArr == maxScore)[0]
        if len(maxScorePlayers) == 1:
            return maxScorePlayers
        else:
            playerBoughtCards = maxScorePlayers.copy()
            for i in range(maxScorePlayers.shape[0]):
                p_idx = maxScorePlayers[i]
                temp_ = 15 * p_idx
                playerBoughtCards[i] = np.sum(env[29 + temp_ : 34 + temp_])

            min_ = np.min(playerBoughtCards)
            winnerIdx = np.where(playerBoughtCards == min_)[0]
            #  if winnerIdx.shape[0] == 1:
            return maxScorePlayers[winnerIdx]
            #  else:
            #      return 4
    else:
        return np.array([-1])


@njit()
def openCard(env, lv1, lv2, lv3, cardId, posE):
    if cardId < 40:
        if lv1[-1] < 40:
            env[posE] = lv1[lv1[-1]]
            lv1[-1] += 1
        else:
            env[posE] = -1
    elif cardId < 70:
        if lv2[-1] < 30:
            env[posE] = lv2[lv2[-1]]
            lv2[-1] += 1
        else:
            env[posE] = -1
    else:
        if lv3[-1] < 20:
            env[posE] = lv3[lv3[-1]]
            lv3[-1] += 1
        else:
            env[posE] = -1


@njit()
def stepEnv(action, env, lv1, lv2, lv3):
    pIdx = env[83] % 4
    temp_ = 15 * pIdx
    pStocks = env[23 + temp_ : 29 + temp_]
    bStocks = env[0:6]
    pPerStocks = env[29 + temp_ : 34 + temp_]
    takenStocks = env[84:89]

    #  Lấy nguyên liệu
    if action < 5:
        takenStocks[action] += 1
        pStocks[action] += 1
        bStocks[action] -= 1

        check_ = False
        s_ = np.sum(takenStocks)
        if s_ == 1:
            if bStocks[action] < 3 and (np.sum(bStocks[0:5]) - bStocks[action]) == 0:
                check_ = True
        elif s_ == 2:
            if (
                np.max(takenStocks) == 2
                or (
                    np.sum(bStocks[0:5])
                    - np.sum(bStocks[np.where(takenStocks == 1)[0]])
                )
                == 0
            ):
                check_ = True
        else:
            check_ = True

        if check_:
            takenStocks[:] = 0

            #  Nếu không thừa nguyên liệu thì next turn
            if np.sum(pStocks) <= 10:
                env[83] += 1

    #  Trả nguyên liệu
    elif action >= 185 and action < 190:
        gem = action - 185
        pStocks[gem] -= 1
        bStocks[gem] += 1

        #  Nếu không thừa nguyên liệu thì next turn
        if np.sum(pStocks) <= 10:
            env[83] += 1

    #  Úp thẻ
    elif (action >= 95 and action < 185) or (action >= 190 and action < 193):
        temp_hideCard = 35 + temp_
        posP = (
            np.where(env[temp_hideCard : temp_hideCard + 3] == -1)[0][0] + temp_hideCard
        )

        if bStocks[5] > 0:
            pStocks[5] += 1
            bStocks[5] -= 1

        if action == 190:  #  Úp thẻ ẩn cấp 1
            env[posP] = lv1[lv1[-1]]
            lv1[-1] += 1
        elif action == 191:  #  Úp thẻ ẩn cấp 2
            env[posP] = lv2[lv2[-1]]
            lv2[-1] += 1
        elif action == 192:  #  Úp thẻ ẩn cấp 3
            env[posP] = lv3[lv3[-1]]
            lv3[-1] += 1
        else:  #  Úp thẻ trên bàn
            cardId = action - 95
            posE = np.where(env[11:23] == cardId)[0][0] + 11
            env[posP] = cardId

            #  Mở thẻ từ chồng úp lên trên bàn chơi
            openCard(env, lv1, lv2, lv3, cardId, posE)

        #  Nếu không thừa nguyên liệu thì next turn
        if np.sum(pStocks) <= 10:
            env[83] += 1

    #  Mua thẻ
    elif action >= 5 and action < 95:
        cardId = action - 5
        if cardId in env[11:23]:
            posE = np.where(env[11:23] == cardId)[0][0] + 11
            inboard = True
        else:
            posE = np.where(env[35 + temp_ : 38 + temp_] == cardId)[0][0] + 35 + temp_
            inboard = False

        cardIn4 = __NORMAL_CARD__[cardId]
        price = cardIn4[6:11]

        nlMat = (price > pPerStocks) * (price - pPerStocks)
        nlBt = np.minimum(nlMat, pStocks[0:5])
        nlG = np.sum(nlMat - nlBt)

        #  Trả nguyên liệu
        pStocks[0:5] -= nlBt  #  Trả nguyên liệu
        pStocks[5] -= nlG
        bStocks[0:5] += nlBt
        bStocks[5] += nlG

        #  Nhận các phần thưởng từ thẻ
        if inboard:
            openCard(env, lv1, lv2, lv3, cardId, posE)
        else:
            env[posE] = -1

        env[34 + temp_] += cardIn4[0]  #  Cộng điểm
        pPerStocks[:] += cardIn4[1:6]

        #  Next turn
        env[83] += 1

    #  40: Bỏ qua lượt (trường hợp đặc biệt khi không thể thực hiện action nào)
    else:
        env[83] += 1

    if (takenStocks == 0).all() and (pPerStocks >= 3).any():
        pos_nobles = np.full(5, 0)
        for i in range(5):
            nobleId = env[6 + i]
            if nobleId != -1:
                nobleIn4 = __NOBLE_CARD__[nobleId]
                price = nobleIn4[1:6]
                if (price <= pPerStocks).all():
                    pos_nobles[i] = 1

        if (pos_nobles == 1).any():
            arr_noble = np.where(pos_nobles == 1)[0]
            if arr_noble.shape[0] == 1:
                choose = 0
            else:
                choose = np.random.randint(0, arr_noble.shape[0])

            noble_idx = arr_noble[choose]
            env[6 + noble_idx] = -1
            env[34 + temp_] += 3


@njit()
def getAgentSize():
    return 4


@njit()
def getStateSize():
    return __STATE_SIZE__


@njit()
def getActionSize():
    return __ACTION_SIZE__


@njit()
def getReward(state):
    if state[263] == 0:
        return -1
    else:
        scoreArr = state[np.array([212, 224, 236, 248])]
        maxScore = np.max(scoreArr)
        if (
            scoreArr[0] < maxScore or scoreArr[0] < 15
        ):  #  Điểm của bản thân không cao nhất
            return 0
        else:
            maxScorePlayers = np.where(scoreArr == maxScore)[0]
            if (
                maxScorePlayers.shape[0] == 1
            ):  #  Bản thân là người duy nhất đạt điểm cao nhất
                return 1
            else:
                playerBoughtCards = maxScorePlayers.copy()
                for i in range(maxScorePlayers.shape[0]):
                    p_idx = maxScorePlayers[i]
                    temp_ = 12 * p_idx
                    playerBoughtCards[i] = np.sum(state[207 + temp_ : 212 + temp_])

                min_ = np.min(playerBoughtCards)
                if playerBoughtCards[0] > min_:  #  Số thẻ của bản thân nhiều hơn
                    return 0
                else:  #  Tất cả đều thắng
                    return 1


@njit()
def one_game_numba(p0, list_other, per_player, per1, per2, per3, p1, p2, p3):
    env, lv1, lv2, lv3 = initEnv()

    winner = np.array([-1])
    while env[83] < 400:
        p_idx = env[83] % 4
        state = getAgentState(env, lv1, lv2, lv3)
        if list_other[p_idx] == -1:
            action, per_player = p0(state, per_player)
            validActions = getValidActions(state)
            if validActions[action] != 1:
                raise Exception("Action không hợp lệ.")
        elif list_other[p_idx] == 1:
            action, per1 = p1(state, per1)
        elif list_other[p_idx] == 2:
            action, per2 = p2(state, per2)
        elif list_other[p_idx] == 3:
            action, per3 = p3(state, per3)
        else:
            raise Exception("Sai list_other.")

        stepEnv(action, env, lv1, lv2, lv3)
        winner = checkEnded(env)
        if winner[0] != -1:
            break

    env[89] = 1
    p0_idx = np.where(list_other == -1)[0][0]
    for p_idx in range(4):
        env[83] = p_idx
        state = getAgentState(env, lv1, lv2, lv3)
        if list_other[p_idx] == -1:
            action, per_player = p0(state, per_player)
        elif list_other[p_idx] == 1:
            action, per1 = p1(state, per1)
        elif list_other[p_idx] == 2:
            action, per2 = p2(state, per2)
        elif list_other[p_idx] == 3:
            action, per3 = p3(state, per3)
        else:
            raise Exception("Sai list_other.")

    if p0_idx in winner:
        result = 1
    else:
        result = 0
    return result, per_player


@njit()
def n_games_numba(p0, num_game, per_player, list_other, per1, per2, per3, p1, p2, p3):
    win = 0
    for _ in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_numba(
            p0, list_other, per_player, per1, per2, per3, p1, p2, p3
        )
        win += winner

    return win, per_player


def one_game_normal(p0, list_other, per_player, per1, per2, per3, p1, p2, p3):
    env, lv1, lv2, lv3 = initEnv()

    winner = np.array([-1])
    while env[83] < 400:
        p_idx = env[83] % 4
        state = getAgentState(env, lv1, lv2, lv3)
        if list_other[p_idx] == -1:
            action, per_player = p0(state, per_player)
            validActions = getValidActions(state)
            if validActions[action] != 1:
                raise Exception("Action không hợp lệ.")
        elif list_other[p_idx] == 1:
            action, per1 = p1(state, per1)
        elif list_other[p_idx] == 2:
            action, per2 = p2(state, per2)
        elif list_other[p_idx] == 3:
            action, per3 = p3(state, per3)
        else:
            raise Exception("Sai list_other.")

        stepEnv(action, env, lv1, lv2, lv3)
        winner = checkEnded(env)
        if winner[0] != -1:
            break

    env[89] = 1
    p0_idx = np.where(list_other == -1)[0][0]
    for p_idx in range(4):
        env[83] = p_idx
        state = getAgentState(env, lv1, lv2, lv3)
        if list_other[p_idx] == -1:
            action, per_player = p0(state, per_player)
        elif list_other[p_idx] == 1:
            action, per1 = p1(state, per1)
        elif list_other[p_idx] == 2:
            action, per2 = p2(state, per2)
        elif list_other[p_idx] == 3:
            action, per3 = p3(state, per3)
        else:
            raise Exception("Sai list_other.")

    if p0_idx in winner:
        result = 1
    else:
        result = 0
    return result, per_player


def n_games_normal(p0, num_game, per_player, list_other, per1, per2, per3, p1, p2, p3):
    win = 0
    for _ in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_normal(
            p0, list_other, per_player, per1, per2, per3, p1, p2, p3
        )
        win += winner

    return win, per_player


import importlib.util
import json, sys

try:
    from env import SHORT_PATH
except:
    pass


def load_module_player(player, game_name=None):
    if game_name == None:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/{player}/Agent_player.py"
        )
    else:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/ifelse/{game_name}/{player}.py"
        )
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


def load_agent(level, *args):
    num_bot = getAgentSize() - 1

    if "_level_" not in globals():
        global _level_
        _level_ = level
        init = True
    else:
        if _level_ != level:
            _level_ = level
            init = True
        else:
            init = False

    if init:
        global _list_per_level_
        global _list_bot_level_
        _list_per_level_ = []
        _list_bot_level_ = []

        if _level_ == 0:
            _list_per_level_ = [
                np.array([[0.0]], dtype=np.float64) for _ in range(num_bot)
            ]
            _list_bot_level_ = [bot_lv0 for _ in range(num_bot)]
        else:
            env_name = sys.argv[1]
            if len(args) > 0:
                dict_level = json.load(
                    open(f"{SHORT_PATH}src/Log/check_system_about_level.json")
                )
            else:
                dict_level = json.load(open(f"{SHORT_PATH}src/Log/level_game.json"))

            if str(_level_) not in dict_level[env_name]:
                raise Exception("Hiện tại không có level này")

            lst_agent_level = dict_level[env_name][str(level)][2]

            for i in range(num_bot):
                if level == -1:
                    module_agent = load_module_player(
                        lst_agent_level[i], game_name=env_name
                    )
                    _list_per_level_.append(module_agent.DataAgent())
                else:
                    data_agent_level = np.load(
                        f"{SHORT_PATH}src/Agent/{lst_agent_level[i]}/Data/{env_name}_{level}/Train.npy",
                        allow_pickle=True,
                    )
                    module_agent = load_module_player(lst_agent_level[i])
                    _list_per_level_.append(
                        module_agent.convert_to_test(data_agent_level)
                    )
                _list_bot_level_.append(module_agent.Test)

    return _list_bot_level_, _list_per_level_


@njit()
def check_run_under_njit(agent, perData):
    return True


def run(
    p0: any = bot_lv0,
    num_game: int = 100,
    per_player: np.ndarray = np.array([[0.0]]),
    level: int = 0,
    *args,
):
    num_bot = getAgentSize() - 1
    list_other = np.array([-1] + [i + 1 for i in range(num_bot)])
    try:
        check_njit = check_run_under_njit(p0, per_player)
    except:
        check_njit = False

    load_agent(level, *args)

    if check_njit:
        return n_games_numba(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
        )
    else:
        return n_games_normal(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
        )
